import asyncio
import json
import logging
import sys
import threading
import time
from pathlib import Path
from typing import Optional

import click
import requests

from tinybird.client import TinyB
from tinybird.tb.modules.cli import cli
from tinybird.tb.modules.common import push_data
from tinybird.tb.modules.datafile.build import folder_build
from tinybird.tb.modules.datafile.fixture import get_fixture_dir
from tinybird.tb.modules.feedback_manager import FeedbackManager
from tinybird.tb.modules.project import Project
from tinybird.tb.modules.shell import Shell, print_table_formatted
from tinybird.tb.modules.watch import watch_project


@cli.command()
@click.option("--watch", is_flag=True, default=False, help="Watch for changes and rebuild automatically")
@click.pass_context
def build(ctx: click.Context, watch: bool) -> None:
    """
    Validate and build the project server side.
    """
    project: Project = ctx.ensure_object(dict)["project"]
    tb_client: TinyB = ctx.ensure_object(dict)["client"]
    click.echo(FeedbackManager.highlight_building_project())

    def process(watch: bool, file_changed: Optional[str] = None, diff: Optional[str] = None) -> None:
        time_start = time.time()
        build_failed = False
        if file_changed and file_changed.endswith(".ndjson"):
            rebuild_fixture(project, tb_client, file_changed)
        else:
            try:
                build_project(project, tb_client, file_changed)
            except click.ClickException as e:
                click.echo(e)
                build_failed = True
            try:
                if file_changed:
                    asyncio.run(folder_build(project, tb_client, filenames=[file_changed]))
                    show_data(tb_client, file_changed, diff)
            except Exception:
                pass

        time_end = time.time()
        elapsed_time = time_end - time_start

        rebuild_str = "Rebuild" if watch else "Build"
        if build_failed:
            click.echo(FeedbackManager.error(message=f"\n✗ {rebuild_str} failed"))
            if not watch:
                sys.exit(1)
        else:
            click.echo(FeedbackManager.success(message=f"\n✓ {rebuild_str} completed in {elapsed_time:.1f}s"))

    process(watch=watch)

    if watch:
        shell = Shell(project=project, tb_client=tb_client)
        click.echo(FeedbackManager.gray(message="\nWatching for changes..."))
        watcher_thread = threading.Thread(
            target=watch_project,
            args=(shell, process, project),
            daemon=True,
        )
        watcher_thread.start()
        shell.run()


def build_project(project: Project, tb_client: TinyB, file_changed: Optional[str] = None) -> None:
    MULTIPART_BOUNDARY_DATA_PROJECT = "data_project://"
    DATAFILE_TYPE_TO_CONTENT_TYPE = {
        ".datasource": "text/plain",
        ".pipe": "text/plain",
    }
    TINYBIRD_API_URL = tb_client.host + "/v1/build"
    logging.debug(TINYBIRD_API_URL)
    TINYBIRD_API_KEY = tb_client.token
    error = False
    try:
        files = [
            ("context://", ("cli-version", "1.0.0", "text/plain")),
        ]
        fds = []
        project_path = project.path
        project_files = project.get_project_files()

        if not project_files:
            return

        for file_path in project_files:
            relative_path = str(Path(file_path).relative_to(project_path))
            fd = open(file_path, "rb")
            fds.append(fd)
            content_type = DATAFILE_TYPE_TO_CONTENT_TYPE.get(Path(file_path).suffix, "application/unknown")
            files.append((MULTIPART_BOUNDARY_DATA_PROJECT, (relative_path, fd.read().decode("utf-8"), content_type)))
        HEADERS = {"Authorization": f"Bearer {TINYBIRD_API_KEY}"}

        r = requests.post(TINYBIRD_API_URL, files=files, headers=HEADERS)
        try:
            result = r.json()
        except Exception as e:
            logging.debug(e, exc_info=True)
            click.echo(FeedbackManager.error(message="Couldn't parse response from server"))
            sys.exit(1)

        logging.debug(json.dumps(result, indent=2))

        build_result = result.get("result")
        if build_result == "success":
            datasources = result.get("datasources", [])
            pipes = result.get("pipes", [])
            if not file_changed:
                for ds in datasources:
                    ds_path_str: Optional[str] = next(
                        (p for p in project_files if p.endswith(ds.get("name") + ".datasource")), None
                    )
                    if ds_path_str:
                        ds_path = Path(ds_path_str)
                        ds_path_str = ds_path_str.replace(f"{project.folder}/", "")
                        click.echo(FeedbackManager.info(message=f"✓ {ds_path_str} created"))
                for pipe in pipes:
                    pipe_name = pipe.get("name")
                    pipe_path_str: Optional[str] = next(
                        (p for p in project_files if p.endswith(pipe_name + ".pipe")), None
                    )
                    if pipe_path_str:
                        pipe_path_str = pipe_path_str.replace(f"{project.folder}/", "")
                        click.echo(FeedbackManager.info(message=f"✓ {pipe_path_str} created"))

            try:
                for filename in project_files:
                    if filename.endswith(".datasource"):
                        ds_path = Path(filename)
                        ds_name = ds_path.stem
                        fixture_folder = get_fixture_dir(project.folder)
                        fixture_path = fixture_folder / f"{ds_name}.ndjson"
                        if fixture_path.exists():
                            append_fixture(tb_client, ds_name, str(fixture_path))

            except Exception as e:
                click.echo(FeedbackManager.error_exception(error=f"Error appending fixtures for '{ds_name}': {e}"))

            feedback = result.get("feedback", [])
            for f in feedback:
                click.echo(
                    FeedbackManager.warning(message=f"△ {f.get('level')}: {f.get('resource')}: {f.get('message')}")
                )

        elif build_result == "failed":
            build_errors = result.get("errors")
            error = True
            for build_error in build_errors:
                filename_bit = f"{build_error.get('filename', '')}"
                error_msg = ((filename_bit + "\n\n") if filename_bit else "") + build_error.get("error")
                click.echo(FeedbackManager.error(message=error_msg))
        else:
            error = True
            click.echo(FeedbackManager.error(message=f"Unknown build result. Error: {result.get('error')}"))

    except Exception as e:
        click.echo(FeedbackManager.error_exception(error="Error: " + str(e)))
    finally:
        for fd in fds:
            fd.close()
        if error:
            raise click.ClickException(FeedbackManager.error(message="Error: Build failed"))


def append_fixture(
    tb_client: TinyB,
    datasource_name: str,
    url: str,
):
    asyncio.run(
        push_data(
            tb_client,
            datasource_name,
            url,
            mode="append",
            concurrency=1,
            silent=True,
        )
    )


def rebuild_fixture(project: Project, tb_client: TinyB, fixture: str) -> None:
    try:
        fixture_path = Path(fixture)
        datasources_path = Path(project.folder) / "datasources"
        ds_name = fixture_path.stem

        if ds_name not in project.datasources:
            try:
                ds_name = "_".join(fixture_path.stem.split("_")[:-1])
            except Exception:
                pass

        ds_path = datasources_path / f"{ds_name}.datasource"

        if ds_path.exists():
            asyncio.run(tb_client.datasource_truncate(ds_name))
            append_fixture(tb_client, ds_name, str(fixture_path))
    except Exception as e:
        click.echo(FeedbackManager.error_exception(error=e))


def show_data(tb_client: TinyB, filename: str, diff: Optional[str] = None):
    table_name = diff
    resource_path = Path(filename)
    resource_name = resource_path.stem

    pipeline = resource_name if filename.endswith(".pipe") else None

    if not table_name:
        table_name = resource_name

    sql = f"SELECT * FROM {table_name} FORMAT JSON"

    res = asyncio.run(tb_client.query(sql, pipeline=pipeline))
    print_table_formatted(res, table_name)
